import functools
import inspect
import logging
from collections import deque
from typing import Any

import pytest as pytest

__all__ = ['AscendPythonReadConnector']


def _get_method(module, name):
  if module and hasattr(module, name):
    meth = getattr(module, name)
    return meth if callable(meth) else None
  return None


def _result_fixture(scope, params):
  @pytest.fixture(scope=scope, params=params)
  def result():
    pass

  return result


class MetadataSupport:
  def __init__(self,
               name,
               fingerprint,
               is_prefix,
               ):
    self.name: str = name
    self.fingerprint: str = fingerprint
    self.is_prefix: bool = is_prefix

  def to_dict(self):
    return dict((k, v) for k, v in self.__dict__.items() if v is not None)


class AscendPythonReadConnector:
  """This decorator automates the testing of read connectors by wrapping the read connector module
  in code that operates as if it was running in the platform"""

  def __init__(self,
               spark=None,
               module=None,
               credentials=None,
               name=None,
               patches=None,
               skip_read_data=False,
               ):
    self.spark_session = spark
    if patches is None:
      patches = []
    self.credentials = credentials
    self.module = module
    self.name = name
    self.def_context = _get_method(module, 'context')
    self.def_list_objects = _get_method(module, 'list_objects')
    self.def_read_bytes = _get_method(module, 'read_bytes')
    self.def_read_spark_dataframe = _get_method(module, 'read_spark_dataframe')
    self.def_read_pandas_dataframe = _get_method(module, 'read_pandas_dataframe')
    self.patches = patches
    self.skip_read_data = skip_read_data

    if not self.spark_session and (self.def_read_spark_dataframe or self.def_read_pandas_dataframe):
      raise ValueError('A spark session is required')
    if not self.def_read_bytes and not self.def_read_spark_dataframe and not self.def_read_pandas_dataframe:
      raise ValueError('At least one read method is required in the module: read_bytes, read_spark_dataframe, read_pandas_dataframe')

    # inject pytest fixtures if they were applied to the global method
    caller_globals = inspect.stack()[1][0].f_globals
    caller_globals['context_result'] = _result_fixture('function', None)
    caller_globals['list_objects_result'] = _result_fixture('function', None)
    caller_globals['read_bytes_result'] = _result_fixture('function', None)
    caller_globals['mock_results'] = _result_fixture('function', None)
    caller_globals['read_dataframe_result'] = _result_fixture('function', None)

  def __process_list_objects(self, context, metadata):
    depth = 0
    queue = deque([(context, metadata, depth)])
    results = []

    while queue:
      current_context, current_metadata, current_depth = queue.popleft()
      for item in self.def_list_objects(current_context, current_metadata):
        logging.debug('partition level: %s meta: %s', current_depth, item)
        results.append(item)
        if item["is_prefix"]:
          queue.append((current_context, item, current_depth + 1))

    return results

  def __call__(self, func):
    @functools.wraps(func)
    def wrapper(*args: Any, **kwargs: Any):
      logging.debug('calling the wrapper')

      # patch as requested
      patched = [m.start() for m in self.patches]
      context_result = self.def_context(self.credentials)
      context_result = context_result if context_result else {}

      # get the list objects/partitions
      list_objects_result = [o for o in self.__process_list_objects(context_result, {})]

      read_data_result = None
      if not self.skip_read_data:
        arguments = {
          'context': context_result,
          'metadata': None,
        }
        if self.def_read_spark_dataframe:
          arguments['spark'] = self.spark_session

        for result in list_objects_result:
          arguments['metadata'] = MetadataSupport(**result).to_dict()
          if self.def_read_bytes:
            assert not self.def_read_spark_dataframe
            assert not self.def_read_pandas_dataframe
            read_data_result = (list(b for b in self.def_read_bytes(**arguments)))
          elif self.def_read_spark_dataframe:
            assert not self.def_read_bytes
            assert not self.def_read_pandas_dataframe
            read_data_result = self.def_read_spark_dataframe(**arguments)
          elif self.def_read_pandas_dataframe:
            assert not self.def_read_bytes
            assert not self.def_read_spark_dataframe
            read_data_result = self.def_read_pandas_dataframe(**arguments)

      logging.debug('read connector interface calls complete')
      # write the data back to the test method for assertion
      full_args = inspect.getfullargspec(func)
      if 'context_result' in full_args.args:
        kwargs['context_result'] = context_result
      if 'list_objects_result' in full_args.args:
        kwargs['list_objects_result'] = list_objects_result
      if 'read_bytes_result' in full_args.args and self.def_read_bytes:
        kwargs['read_bytes_result'] = read_data_result
      if 'mock_results' in full_args.args:
        kwargs['mock_results'] = patched
      if 'read_dataframe_result' in full_args.args and (self.def_read_spark_dataframe or self.def_read_pandas_dataframe):
        kwargs['read_dataframe_result'] = read_data_result

      # un-patch if required
      [p.stop() for p in patched if hasattr(p, "stop")]

      return func(*args, **kwargs)

    return wrapper



import functools
import inspect
import io
import logging
from typing import Any

import pytest

__all__ = ['AscendPythonBytestreamReadConnector']


def _get_method(module, name):
  if module and hasattr(module, name):
    meth = getattr(module, name)
    return meth if callable(meth) else None
  return None


def _result_fixture(scope, params):
  @pytest.fixture(scope=scope, params=params)
  def result():
    pass

  return result


class AscendPythonBytestreamReadConnector:
  """Read as if processing bytestream from S3 connectors"""

  def __init__(self,
               module=None,
               bytestream=None,
               reader=None,
               ):
    self.module = module
    self.bytestream = bytes(bytestream, 'utf-8') if isinstance(bytestream, str) else bytes(bytestream) if bytestream else None
    self.reader = reader if reader else io.BytesIO(self.bytestream)
    self.def_parser_function = _get_method(self.module, 'parser_function')
    self.on_next = []

    # inject pytest fixtures if they were applied to the global method
    caller_globals = inspect.stack()[1][0].f_globals
    caller_globals['on_next_results'] = _result_fixture('function', None)

  def _on_next(self, record):
    self.on_next.append(record)

  def __call__(self, func):
    @functools.wraps(func)
    def wrapper(*args: Any, **kwargs: Any):
      logging.debug('calling the wrapper')

      self.def_parser_function(self.reader, self._on_next)

      logging.debug('read connector interface calls complete')
      # write the data back to the test method for assertion
      full_args = inspect.getfullargspec(func)
      if 'on_next_results' in full_args.args:
        kwargs['on_next_results'] = self.on_next

      return func(*args, **kwargs)

    return wrapper



import decimal
import functools
import inspect
import logging
from datetime import datetime
from typing import Any

import pyspark.sql.types
import pyspark.sql.types as T
import pytest as pytest
from pyspark.sql import SparkSession, DataFrame

__all__ = ['AscendPySparkTransform']


def _get_method(module, name):
  if module and hasattr(module, name):
    meth = getattr(module, name)
    return meth if callable(meth) else None
  return None


def _result_fixture(scope, params):
  @pytest.fixture(scope=scope, params=params)
  def result():
    pass

  return result


def _get_schema_discovery_df(spark: SparkSession, schema: T.StructType) -> DataFrame:
  """Capture a dataframe with default values for the purposes of determining
  the output schema for a particular transformation"""
  now = datetime.now()

  def get_dummy_val(dt: T.DataType) -> Any:
    if isinstance(dt, T.StringType):
      return 'A'
    if isinstance(dt, T.BooleanType):
      return False
    if isinstance(dt, (T.DoubleType, T.FloatType)):
      return 1.0
    if isinstance(dt, (T.IntegerType, T.LongType, T.ShortType)):
      return 1
    if isinstance(dt, T.DecimalType):
      return decimal.Decimal(0)
    if isinstance(dt, T.DateType):
      return now.date()
    if isinstance(dt, T.TimestampType):
      return now
    return None

  row = {f.name: get_dummy_val(f.dataType) for f in schema}
  return spark.createDataFrame(data=[T.Row(**row)], schema=schema, verifySchema=False)


class AscendPySparkTransform:
  """This decorator automates the testing of custom python transforms."""

  def __init__(self,
               spark,
               module=None,
               discover_schema=True,
               schema=None,
               data=None,
               patches=None,
               credentials=None,
               ):
    if data is None:
      data = []
    if patches is None:
      patches = []
    self.spark_session = spark
    self.module = module
    self.discover_schema = discover_schema
    self.in_schema = schema
    self.in_data = data
    self.credentials = credentials
    self.patches = patches
    self.def_transform = _get_method(self.module, 'transform')
    self.def_infer_schema = _get_method(self.module, 'infer_schema')

    caller_globals = inspect.stack()[1][0].f_globals
    caller_globals['transform_schema'] = _result_fixture('function', None)
    caller_globals['transform_exception'] = _result_fixture('function', None)
    caller_globals['input_dataframe'] = _result_fixture('function', None)
    caller_globals['transform_result'] = _result_fixture('function', None)
    caller_globals['mock_results'] = _result_fixture('function', None)

  def __call__(self, func):
    @functools.wraps(func)
    def wrapper(*args: Any, **kwargs: Any):
      logging.debug('calling the transform wrapper')

      # patch as requested
      patched = [m.start() for m in self.patches if hasattr(m, "start")]

      transform_schema_df = None

      if self.discover_schema:
        if self.def_infer_schema:
          transform_schema = self.def_infer_schema(self.spark_session, [], self.credentials)
          transform_schema_df = self.spark_session.createDataFrame(schema=transform_schema, data=[])
        else:
          empty_df = _get_schema_discovery_df(self.spark_session, self.in_schema)
          transform_schema_df = self.def_transform(self.spark_session, [empty_df], self.credentials)
          logging.debug(f'schema discovery returned {transform_schema_df}')

      if self.in_data:
        input_dataframe = self.spark_session.createDataFrame(self.in_data, self.in_schema)
      else:
        input_dataframe = self.spark_session.createDataFrame(self.spark_session.sparkContext.emptyRDD(),
                                                             self.in_schema if self.in_schema else pyspark.sql.types.StructType())

      transform_result = None
      transform_exception = None
      try:
        transform_result = self.def_transform(self.spark_session, [input_dataframe], self.credentials)
      except Exception as ex:
        transform_exception = ex

      logging.debug('transform interface calls complete')
      # write the data back to the test method for assertion
      full_args = inspect.getfullargspec(func)
      if 'transform_schema' in full_args.args:
        kwargs['transform_schema'] = transform_schema_df.schema if transform_schema_df else None
      if 'input_dataframe' in full_args.args:
        kwargs['input_dataframe'] = input_dataframe
      if 'transform_result' in full_args.args:
        kwargs['transform_result'] = transform_result
      if 'mock_results' in full_args.args:
        kwargs['mock_results'] = patched
      if 'transform_exception' in full_args.args:
        kwargs['transform_exception'] = transform_exception
      elif transform_exception:
        raise transform_exception

      # un-patch if required
      [p.stop() for p in self.patches if hasattr(p, "stop")]

      return func(*args, **kwargs)

    return wrapper



import logging


from .ascend_pyspark_transform import *
from .ascend_python_bytestream_read_connector import *
from .ascend_python_read_connector import *

__all__ = ['AscendPySparkTransform',
           'AscendPythonBytestreamReadConnector',
           'AscendPythonReadConnector', ]

logging.getLogger().setLevel(logging.WARN)